@@ -1,20 +1,22 @@
-import os
-import shutil
-import pathlib
 import importlib
+import os
+import pathlib
+import shutil
 import subprocess
+import sys
 
 import click
 import spin
 from spin.cmds import meson
 
+IS_PYPY = (sys.implementation.name == 'pypy')
 
 # Check that the meson git submodule is present
 curdir = pathlib.Path(__file__).parent
 meson_import_dir = curdir.parent / 'vendored-meson' / 'meson' / 'mesonbuild'
 if not meson_import_dir.exists():
     raise RuntimeError(
-        'The `vendored-meson/meson` git submodule does not exist! ' +
+        'The `vendored-meson/meson` git submodule does not exist! '
         'Run `git submodule update --init` to fix this problem.'
     )
 
@@ -46,8 +48,8 @@ def changelog(token, revision_range):
     $ spin authors -t $GH_TOKEN --revision-range v1.25.0..v1.26.0
     """
     try:
-        from github.GithubException import GithubException
         from git.exc import GitError
+        from github.GithubException import GithubException
         changelog = _get_numpy_tools(pathlib.Path('changelog.py'))
     except ModuleNotFoundError as e:
         raise click.ClickException(
@@ -127,19 +129,36 @@ def docs(*, parent_callback, **kwargs):
 jobs_param = next(p for p in docs.params if p.name == 'jobs')
 jobs_param.default = 1
 
+if IS_PYPY:
+    default = "not slow and not slow_pypy"
+else:
+    default = "not slow"
 
 @click.option(
     "-m",
     "markexpr",
     metavar='MARKEXPR',
-    default="not slow",
+    default=default,
     help="Run tests with the given markers"
 )
+@click.option(
+    "-p",
+    "--parallel-threads",
+    metavar='PARALLEL_THREADS',
+    default="1",
+    help="Run tests many times in number of parallel threads under pytest-run-parallel."
+         " Can be set to `auto` to use all cores. Use `spin test -p <number> -- "
+         "--skip-thread-unsafe=true` to only run tests that can run in parallel. "
+         "pytest-run-parallel must be installed to use."
+)
 @spin.util.extend_command(spin.cmds.meson.test)
-def test(*, parent_callback, pytest_args, tests, markexpr, **kwargs):
+def test(*, parent_callback, pytest_args, tests, markexpr, parallel_threads, **kwargs):
     """
     By default, spin will run `-m 'not slow'`. To run the full test suite, use
     `spin test -m full`
+
+    When pytest-run-parallel is avaliable, use `spin test -p auto` or
+    `spin test -p <num_of_threads>` to run tests sequentional in parallel threads.
     """  # noqa: E501
     if (not pytest_args) and (not tests):
         pytest_args = ('--pyargs', 'numpy')
@@ -148,6 +167,9 @@ def test(*, parent_callback, pytest_args, tests, markexpr, **kwargs):
         if markexpr != "full":
             pytest_args = ('-m', markexpr) + pytest_args
 
+    if parallel_threads != "1":
+        pytest_args = ('--parallel-threads', parallel_threads) + pytest_args
+
     kwargs['pytest_args'] = pytest_args
     parent_callback(**{'pytest_args': pytest_args, 'tests': tests, **kwargs})
 
@@ -187,9 +209,11 @@ def check_docs(*, parent_callback, pytest_args, **kwargs):
     """  # noqa: E501
     try:
         # prevent obscure error later
-        import scipy_doctest
+        import scipy_doctest  # noqa: F401
     except ModuleNotFoundError as e:
         raise ModuleNotFoundError("scipy-doctest not installed") from e
+    if scipy_doctest.__version__ < '1.8.0':
+        raise ModuleNotFoundError("please update scipy_doctests to >= 1.8.0")
 
     if (not pytest_args):
         pytest_args = ('--pyargs', 'numpy')
@@ -197,6 +221,7 @@ def check_docs(*, parent_callback, pytest_args, **kwargs):
     # turn doctesting on:
     doctest_args = (
         '--doctest-modules',
+        '--doctest-only-doctests=true',
         '--doctest-collect=api'
     )
 
@@ -257,6 +282,7 @@ def _set_mem_rlimit(max_mem=None):
     Set address space rlimit
     """
     import resource
+
     import psutil
 
     mem = psutil.virtual_memory()
@@ -277,7 +303,7 @@ def _set_mem_rlimit(max_mem=None):
 def _commit_to_sha(commit):
     p = spin.util.run(['git', 'rev-parse', commit], output=False, echo=False)
     if p.returncode != 0:
-        raise(
+        raise (
             click.ClickException(
                 f'Could not find SHA matching commit `{commit}`'
             )
@@ -320,36 +346,20 @@ def _run_asv(cmd):
 
 @click.command()
 @click.option(
-    "-b", "--branch",
-    metavar='branch',
-    default="main",
-)
-@click.option(
-    '--uncommitted',
+    '--fix',
     is_flag=True,
     default=False,
     required=False,
 )
 @click.pass_context
-def lint(ctx, branch, uncommitted):
-    """🔦 Run lint checks on diffs.
-    Provide target branch name or `uncommitted` to check changes before committing:
+def lint(ctx, fix):
+    """🔦 Run lint checks with Ruff
 
     \b
-    Examples:
+    To run automatic fixes use:
 
     \b
-    For lint checks of your development branch with `main` or a custom branch:
-
-    \b
-    $ spin lint # defaults to main
-    $ spin lint --branch custom_branch
-
-    \b
-    To check just the uncommitted changes before committing
-
-    \b
-    $ spin lint --uncommitted
+    $ spin lint --fix
     """
     try:
         linter = _get_numpy_tools(pathlib.Path('linter.py'))
@@ -358,7 +368,7 @@ def lint(ctx, branch, uncommitted):
             f"{e.msg}. Install using requirements/linter_requirements.txt"
         )
 
-    linter.DiffLinter(branch).run_lint(uncommitted)
+    linter.DiffLinter().run_lint(fix)
 
 @click.command()
 @click.option(
@@ -382,6 +392,11 @@ def lint(ctx, branch, uncommitted):
     '--quick', '-q', is_flag=True, default=False,
     help="Run each benchmark only once (timings won't be accurate)"
 )
+@click.option(
+    '--factor', '-f', default=1.05,
+    help="The factor above or below which a benchmark result is "
+         "considered reportable. This is passed on to the asv command."
+)
 @click.argument(
     'commits', metavar='',
     required=False,
@@ -389,7 +404,7 @@ def lint(ctx, branch, uncommitted):
 )
 @meson.build_dir_option
 @click.pass_context
-def bench(ctx, tests, compare, verbose, quick, commits, build_dir):
+def bench(ctx, tests, compare, verbose, quick, factor, commits, build_dir):
     """🏋 Run benchmarks.
 
     \b
@@ -444,7 +459,7 @@ def bench(ctx, tests, compare, verbose, quick, commits, build_dir):
         meson._set_pythonpath(build_dir)
 
         p = spin.util.run(
-            ['python', '-c', 'import numpy as np; print(np.__version__)'],
+            [sys.executable, '-c', 'import numpy as np; print(np.__version__)'],
             cwd='benchmarks',
             echo=False,
             output=False
@@ -461,7 +476,7 @@ def bench(ctx, tests, compare, verbose, quick, commits, build_dir):
         ] + bench_args
         _run_asv(cmd)
     else:
-        # Ensure that we don't have uncommited changes
+        # Ensure that we don't have uncommitted changes
         commit_a, commit_b = [_commit_to_sha(c) for c in commits]
 
         if commit_b == 'HEAD' and _dirty_git_working_dir():
@@ -472,7 +487,7 @@ def bench(ctx, tests, compare, verbose, quick, commits, build_dir):
             )
 
         cmd_compare = [
-            'asv', 'continuous', '--factor', '1.05',
+            'asv', 'continuous', '--factor', str(factor),
         ] + bench_args + [commit_a, commit_b]
         _run_asv(cmd_compare)
 
@@ -527,6 +542,48 @@ def mypy(ctx):
     ctx.forward(test)
 
 
+@click.command()
+@click.option(
+    '--concise',
+    is_flag=True,
+    default=False,
+    help="Concise output format",
+)
+@meson.build_dir_option
+def stubtest(*, concise: bool, build_dir: str) -> None:
+    """🧐 Run stubtest on NumPy's .pyi stubs
+
+    Requires mypy to be installed
+    """
+    click.get_current_context().invoke(build)
+    meson._set_pythonpath(build_dir)
+    print(f"{build_dir = !r}")
+
+    import sysconfig
+    purellib = sysconfig.get_paths()["purelib"]
+    print(f"{purellib = !r}")
+
+    stubtest_dir = curdir.parent / 'tools' / 'stubtest'
+    mypy_config = stubtest_dir / 'mypy.ini'
+    allowlists = [stubtest_dir / 'allowlist.txt']
+    if sys.version_info < (3, 12):
+        allowlists.append(stubtest_dir / 'allowlist_py311.txt')
+    else:
+        allowlists.append(stubtest_dir / 'allowlist_py312.txt')
+
+    cmd = [
+        'stubtest',
+        '--ignore-disjoint-bases',
+        f'--mypy-config-file={mypy_config}',
+        *(f'--allowlist={allowlist}' for allowlist in allowlists),
+    ]
+    if concise:
+        cmd.append('--concise')
+    cmd.append('numpy')
+
+    spin.util.run(cmd)
+
+
 @click.command(context_settings={
     'ignore_unknown_options': True
 })
@@ -625,7 +682,8 @@ def notes(version_override):
     )
 
     try:
-        test_notes = _get_numpy_tools(pathlib.Path('ci', 'test_all_newsfragments_used.py'))
+        cmd = pathlib.Path('ci', 'test_all_newsfragments_used.py')
+        test_notes = _get_numpy_tools(cmd)
     except ModuleNotFoundError as e:
         raise click.ClickException(
             f"{e.msg}. Install the missing packages to use this command."
